/**
 * eobjects.org MetaModel
 * Copyright (C) 2010 eobjects.org
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution; if not, write to:
 * Free Software Foundation, Inc.
 * 51 Franklin Street, Fifth Floor
 * Boston, MA  02110-1301  USA
 */
package org.eobjects.metamodel;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.eobjects.metamodel.data.DataSet;
import org.eobjects.metamodel.data.EmptyDataSet;
import org.eobjects.metamodel.data.FirstRowDataSet;
import org.eobjects.metamodel.data.InMemoryDataSet;
import org.eobjects.metamodel.data.MaxRowsDataSet;
import org.eobjects.metamodel.data.Row;
import org.eobjects.metamodel.detect.ColumnTypeTransformer;
import org.eobjects.metamodel.detect.ColumnTypeTransformerDataSet;
import org.eobjects.metamodel.query.FilterItem;
import org.eobjects.metamodel.query.FromItem;
import org.eobjects.metamodel.query.GroupByItem;
import org.eobjects.metamodel.query.JoinType;
import org.eobjects.metamodel.query.OperatorType;
import org.eobjects.metamodel.query.OrderByItem;
import org.eobjects.metamodel.query.Query;
import org.eobjects.metamodel.query.SelectItem;
import org.eobjects.metamodel.schema.Column;
import org.eobjects.metamodel.schema.ColumnType;
import org.eobjects.metamodel.schema.MutableColumn;
import org.eobjects.metamodel.schema.MutableRelationship;
import org.eobjects.metamodel.schema.MutableSchema;
import org.eobjects.metamodel.schema.MutableTable;
import org.eobjects.metamodel.schema.Relationship;
import org.eobjects.metamodel.schema.Schema;
import org.eobjects.metamodel.schema.Table;
import org.eobjects.metamodel.schema.TableType;
import org.eobjects.metamodel.util.CollectionUtils;

/**
 * Abstract DataContext for data sources that do not support SQL queries
 * natively.
 * 
 * Instead this superclass only requires that a subclass can materialize a
 * single table at a time. Then the query will be executed by post processing
 * the datasets client-side.
 */
public abstract class QueryPostprocessDataContext extends AbstractDataContext {

	private static final Logger logger = LoggerFactory
			.getLogger(QueryPostprocessDataContext.class);

	public static final String INFORMATION_SCHEMA_NAME = "information_schema";

	private final Map<String, ColumnTypeTransformer> _columnTypeTransformers = new HashMap<String, ColumnTypeTransformer>();
	private String[] _schemaNames;
	private Schema _mainSchema;

	public void setColumnTypeTransformer(String qualifiedColumnName,
			ColumnTypeTransformer transformer) {
		_columnTypeTransformers.put(qualifiedColumnName, transformer);
	}

	public void setColumnTypeTransformer(Column column,
			ColumnTypeTransformer transformer) {
		_columnTypeTransformers.put(column.getQualifiedLabel(), transformer);
	}

	@Override
	public DataSet executeQuery(final Query query) {
		final List<SelectItem> selectItems = query.getSelectClause().getItems();
		final List<FromItem> fromItems = query.getFromClause().getItems();
		final List<FilterItem> whereItems = query.getWhereClause().getItems();
		final List<SelectItem> whereSelectItems = query.getWhereClause()
				.getEvaluatedSelectItems();
		final List<GroupByItem> groupByItems = query.getGroupByClause()
				.getItems();
		final List<SelectItem> havingSelectItems = query.getHavingClause()
				.getEvaluatedSelectItems();
		final List<FilterItem> havingItems = query.getHavingClause().getItems();
		final List<OrderByItem> orderByItems = query.getOrderByClause()
				.getItems();

		// Check for very simple queries with max rows property set (typically
		// preview), see Ticket #187
		final Integer maxRows = query.getMaxRows();
		final Integer firstRow = query.getFirstRow();

		previewTable: if (maxRows != null && whereItems.isEmpty()
				&& groupByItems.isEmpty() && havingItems.isEmpty()
				&& orderByItems.isEmpty() && fromItems.size() == 1) {

			// rows to materialize - depending on the First row property
			final int rowsToMaterialize;
			if (firstRow == null) {
				rowsToMaterialize = maxRows;
			} else {
				rowsToMaterialize = maxRows + firstRow;
			}

			final Table table = fromItems.get(0).getTable();
			if (table != null) {
				Column[] columns = new Column[selectItems.size()];
				int i = 0;
				for (Iterator<SelectItem> it = selectItems.iterator(); it
						.hasNext();) {
					SelectItem item = it.next();
					if (item.getFunction() != null
							|| item.getExpression() != null) {
						break previewTable;
					}
					columns[i] = item.getColumn();
					i++;
				}

				DataSet dataSet = materializeTable(table, columns,
						rowsToMaterialize);
				dataSet = MetaModelHelper.getSelection(selectItems, dataSet);

				if (firstRow != null) {
					dataSet = new FirstRowDataSet(dataSet, firstRow);
				}
				return dataSet;
			}
		}

		// Materialize the tables in the from clause
		final DataSet[] fromDataSets = new DataSet[fromItems.size()];
		for (int i = 0; i < fromDataSets.length; i++) {
			FromItem fromItem = fromItems.get(i);
			fromDataSets[i] = materializeFromItem(fromItem);
		}

		// Creates a list for all select items that are needed to execute query
		// (some may only be used as part of a filter, but not shown in result)
		final List<SelectItem> workSelectItems = CollectionUtils.concat(
				selectItems, whereSelectItems, havingSelectItems, true);

		// Execute the query using the raw data
		DataSet dataSet = MetaModelHelper.getCarthesianProduct(fromDataSets,
				whereItems);
		if (groupByItems.size() > 0) {
			dataSet = MetaModelHelper.getGrouped(workSelectItems, dataSet,
					groupByItems);
		} else {
			dataSet = MetaModelHelper.getAggregated(workSelectItems, dataSet);
		}
		dataSet = MetaModelHelper.getFiltered(dataSet, havingItems);

		if (query.getSelectClause().isDistinct()) {
			dataSet = MetaModelHelper.getSelection(selectItems, dataSet);
			dataSet = MetaModelHelper.getDistinct(dataSet);
			dataSet = MetaModelHelper.getOrdered(dataSet, orderByItems);
		} else {
			dataSet = MetaModelHelper.getOrdered(dataSet, orderByItems);
			dataSet = MetaModelHelper.getSelection(selectItems, dataSet);
		}

		if (firstRow != null) {
			dataSet = new FirstRowDataSet(dataSet, firstRow);
		}
		if (maxRows != null) {
			dataSet = new MaxRowsDataSet(dataSet, maxRows);
		}
		return dataSet;
	}

	public QueryPostprocessDataContext transformColumnTypes(Table table) {
		return transformColumnTypes(table.getColumnsOfType(ColumnType.VARCHAR));
	}

	public QueryPostprocessDataContext transformColumnTypes(Column... columns) {
		Table[] tables = MetaModelHelper.getTables(columns);
		for (Table table : tables) {
			Column[] columnsToQuery = MetaModelHelper.getTableColumns(table,
					columns);
			columnsToQuery = MetaModelHelper.getColumnsByType(columnsToQuery,
					ColumnType.VARCHAR);
			DataSet dataSet = executeQuery(new Query().select(columnsToQuery)
					.from(table));

			for (Column column : columnsToQuery) {
				assert column instanceof MutableColumn;
				ColumnTypeTransformer transformer = new ColumnTypeTransformer(
						(MutableColumn) column);
				_columnTypeTransformers.put(column.getQualifiedLabel(),
						transformer);
			}
			while (dataSet.next()) {
				Row row = dataSet.getRow();
				for (Column column : columnsToQuery) {
					ColumnTypeTransformer transformer = _columnTypeTransformers
							.get(column.getQualifiedLabel());
					Object value = row.getValue(column);
					if (value instanceof String || value == null) {
						transformer.registerValue((String) value);
					}
				}
			}
			for (Column column : columnsToQuery) {
				ColumnTypeTransformer transformer = _columnTypeTransformers
						.get(column.getQualifiedLabel());
				transformer.narrowColumnType();
				if (column.getType() == ColumnType.VARCHAR) {
					_columnTypeTransformers.remove(column.getQualifiedLabel());
				}
			}
		}
		return this;
	}

	/**
	 * Performs column type transformation for the whole schema
	 */
	public QueryPostprocessDataContext autoTransformColumnTypes() {
		Schema schema = getMainSchemaInternal();
		Table[] tables = schema.getTables();
		for (Table table : tables) {
			transformColumnTypes(table);
		}
		return this;
	}

	protected DataSet materializeFromItem(FromItem fromItem) {
		DataSet dataSet;
		JoinType joinType = fromItem.getJoin();
		if (fromItem.getTable() != null) {
			// We need to materialize a single table
			Table table = fromItem.getTable();
			Column[] columns = table.getColumns();
			Query query = fromItem.getQuery();
			if (query != null) {
				List<SelectItem> selectItems = CollectionUtils.concat(query
						.getSelectClause().getItems(), query.getWhereClause()
						.getEvaluatedSelectItems(), query.getHavingClause()
						.getEvaluatedSelectItems(), true);

				List<Column> colsToMaterialize = new ArrayList<Column>();
				for (SelectItem selectItem : selectItems) {
					Column column = selectItem.getColumn();
					if (column != null && column.getTable() != null
							&& column.getTable().equals(table)) {
						colsToMaterialize.add(column);
					}
				}
				selectItems = null;
				columns = colsToMaterialize
						.toArray(new Column[colsToMaterialize.size()]);
			}
			// Dispatching to the concrete subclass of
			// QueryPostprocessDataContextStrategy
			if (logger.isDebugEnabled()) {
				logger.debug("calling materializeTable(" + table.getName()
						+ "," + Arrays.toString(columns) + ",-1");
			}
			dataSet = materializeTable(table, columns, -1);
		} else if (joinType != null) {
			// We need to (recursively) materialize a joined FromItem
			if (fromItem.getLeftSide() == null
					|| fromItem.getRightSide() == null) {
				throw new IllegalArgumentException(
						"Joined FromItem requires both left and right side: "
								+ fromItem);
			}
			DataSet[] fromItemDataSets = new DataSet[2];
			fromItemDataSets[0] = materializeFromItem(fromItem.getLeftSide());
			fromItemDataSets[1] = materializeFromItem(fromItem.getRightSide());
			SelectItem[] leftOn = fromItem.getLeftOn();
			SelectItem[] rightOn = fromItem.getRightOn();
			FilterItem[] onConditions = new FilterItem[leftOn.length];
			for (int i = 0; i < onConditions.length; i++) {
				FilterItem whereItem = new FilterItem(leftOn[i],
						OperatorType.EQUALS_TO, rightOn[i]);
				onConditions[i] = whereItem;
			}
			if (joinType == JoinType.INNER) {
				dataSet = MetaModelHelper.getCarthesianProduct(
						fromItemDataSets, onConditions);
			} else if (joinType == JoinType.LEFT) {
				dataSet = MetaModelHelper.getLeftJoin(fromItemDataSets[0],
						fromItemDataSets[1], onConditions);
			} else if (joinType == JoinType.RIGHT) {
				dataSet = MetaModelHelper.getRightJoin(fromItemDataSets[0],
						fromItemDataSets[1], onConditions);
			} else {
				throw new IllegalArgumentException(
						"FromItem type not supported: " + fromItem);
			}
		} else if (fromItem.getSubQuery() != null) {
			// We need to (recursively) materialize a subquery
			dataSet = executeQuery(fromItem.getSubQuery());
		} else {
			throw new IllegalArgumentException("FromItem type not supported: "
					+ fromItem);
		}
		if (dataSet == null) {
			throw new IllegalStateException(
					"FromItem was not succesfully materialized: " + fromItem);
		}
		return dataSet;
	}

	protected DataSet materializeTable(Table table, Column[] columns,
			int maxRows) {
		if (columns == null || columns.length == 0) {
			columns = table.getColumns();
		}

		if (maxRows == 0) {
			return new EmptyDataSet(columns);
		}

		if (table != null) {
			Schema schema = table.getSchema();
			if (INFORMATION_SCHEMA_NAME.equals(schema.getName())) {
				return materializeInformationSchemaTable(table, columns,
						maxRows);
			} else {
				DataSet tableDataSet = materializeMainSchemaTable(table,
						columns, maxRows);
				if (_columnTypeTransformers.isEmpty()) {
					return tableDataSet;
				} else {
					DataSet transformedDataSet = new ColumnTypeTransformerDataSet(
							tableDataSet, _columnTypeTransformers);
					return transformedDataSet;
				}
			}
		}
		return null;
	}

	protected boolean isMainSchemaTable(Table table) {
		Schema schema = table.getSchema();
		if (INFORMATION_SCHEMA_NAME.equals(schema.getName())) {
			return false;
		} else {
			return true;
		}
	}

	@Override
	protected final String[] getSchemaNamesInternal() throws MetaModelException {
		if (_schemaNames == null) {
			_schemaNames = new String[2];
			_schemaNames[0] = INFORMATION_SCHEMA_NAME;
			_schemaNames[1] = getMainSchemaName();
		}
		return Arrays.copyOf(_schemaNames, _schemaNames.length);
	}

	@Override
	protected String getDefaultSchemaName() throws MetaModelException {
		return getSchemaNames()[1];
	}

	@Override
	protected final Schema getSchemaByNameInternal(String name)
			throws MetaModelException {
		if (name != null) {
			if (name.equals(getSchemaNames()[1])) {
				return getMainSchemaInternal();
			} else if (name.equals(INFORMATION_SCHEMA_NAME)) {
				return getInformationSchema();
			}
		}
		return null;
	}

	private Schema getInformationSchema() {
		// Create schema
		MutableSchema informationSchema = new MutableSchema(
				INFORMATION_SCHEMA_NAME);
		MutableTable tablesTable = new MutableTable("tables", TableType.TABLE,
				informationSchema);
		MutableTable columnsTable = new MutableTable("columns",
				TableType.TABLE, informationSchema);
		MutableTable relationshipsTable = new MutableTable("relationships",
				TableType.TABLE, informationSchema);
		informationSchema.addTable(tablesTable).addTable(columnsTable)
				.addTable(relationshipsTable);

		// Create "tables" table: name, type, num_columns, remarks
		tablesTable.addColumn(new MutableColumn("name", ColumnType.VARCHAR,
				tablesTable, 0, false));
		tablesTable.addColumn(new MutableColumn("type", ColumnType.VARCHAR,
				tablesTable, 1, true));
		tablesTable.addColumn(new MutableColumn("num_columns",
				ColumnType.INTEGER, tablesTable, 2, true));
		tablesTable.addColumn(new MutableColumn("remarks", ColumnType.VARCHAR,
				tablesTable, 3, true));

		// Create "columns" table: name, type, native_type, size, nullable,
		// indexed, table, remarks
		columnsTable.addColumn(new MutableColumn("name", ColumnType.VARCHAR,
				columnsTable, 0, false));
		columnsTable.addColumn(new MutableColumn("type", ColumnType.VARCHAR,
				columnsTable, 1, true));
		columnsTable.addColumn(new MutableColumn("native_type",
				ColumnType.VARCHAR, columnsTable, 2, true));
		columnsTable.addColumn(new MutableColumn("size", ColumnType.INTEGER,
				columnsTable, 3, true));
		columnsTable.addColumn(new MutableColumn("nullable",
				ColumnType.BOOLEAN, columnsTable, 4, true));
		columnsTable.addColumn(new MutableColumn("indexed", ColumnType.BOOLEAN,
				columnsTable, 5, true));
		columnsTable.addColumn(new MutableColumn("table", ColumnType.VARCHAR,
				columnsTable, 6, false));
		columnsTable.addColumn(new MutableColumn("remarks", ColumnType.VARCHAR,
				columnsTable, 7, true));

		// Create "relationships" table: primary_table, primary_column,
		// foreign_table, foreign_column
		relationshipsTable.addColumn(new MutableColumn("primary_table",
				ColumnType.VARCHAR, relationshipsTable, 0, false));
		relationshipsTable.addColumn(new MutableColumn("primary_column",
				ColumnType.VARCHAR, relationshipsTable, 1, false));
		relationshipsTable.addColumn(new MutableColumn("foreign_table",
				ColumnType.VARCHAR, relationshipsTable, 2, false));
		relationshipsTable.addColumn(new MutableColumn("foreign_column",
				ColumnType.VARCHAR, relationshipsTable, 3, false));

		MutableRelationship.createRelationship(
				tablesTable.getColumnByName("name"),
				columnsTable.getColumnByName("table"));
		MutableRelationship.createRelationship(
				tablesTable.getColumnByName("name"),
				relationshipsTable.getColumnByName("primary_table"));
		MutableRelationship.createRelationship(
				tablesTable.getColumnByName("name"),
				relationshipsTable.getColumnByName("foreign_table"));
		MutableRelationship.createRelationship(
				columnsTable.getColumnByName("name"),
				relationshipsTable.getColumnByName("primary_column"));
		MutableRelationship.createRelationship(
				columnsTable.getColumnByName("name"),
				relationshipsTable.getColumnByName("foreign_column"));

		return informationSchema;
	}

	private DataSet materializeInformationSchemaTable(Table table,
			Column[] columns, int maxRows) {
		DataSet dataSet = null;
		String tableName = table.getName();
		if ("tables".equals(tableName)) {
			// "tables" columns: name, type, num_columns, remarks
			List<Object[]> data = new ArrayList<Object[]>();
			for (Table t : getMainSchemaInternal().getTables()) {
				String typeString = null;
				if (t.getType() != null) {
					typeString = t.getType().toString();
				}
				data.add(new Object[] { t.getName(), typeString,
						t.getColumnCount(), t.getRemarks() });
			}

			dataSet = new InMemoryDataSet(
					MetaModelHelper.createSelectItems(table.getColumns()), data);
		} else if ("columns".equals(tableName)) {
			// "columns" columns: name, type, native_type, size, nullable,
			// indexed, table, remarks
			List<Object[]> data = new ArrayList<Object[]>();
			for (Table t : getMainSchemaInternal().getTables()) {
				for (Column c : t.getColumns()) {
					String typeString = null;
					if (t.getType() != null) {
						typeString = c.getType().toString();
					}
					data.add(new Object[] { c.getName(), typeString,
							c.getNativeType(), c.getColumnSize(),
							c.isNullable(), c.isIndexed(), t.getName(),
							c.getRemarks() });
				}
			}
			dataSet = new InMemoryDataSet(
					MetaModelHelper.createSelectItems(table.getColumns()), data);
		} else if ("relationships".equals(tableName)) {
			// "relationships" columns: primary_table, primary_column,
			// foreign_table, foreign_column
			List<Object[]> data = new ArrayList<Object[]>();
			for (Relationship r : getMainSchemaInternal().getRelationships()) {
				Column[] primaryColumns = r.getPrimaryColumns();
				Column[] foreignColumns = r.getForeignColumns();
				Table pTable = r.getPrimaryTable();
				Table fTable = r.getForeignTable();
				for (int i = 0; i < primaryColumns.length; i++) {
					Column pColumn = primaryColumns[i];
					Column fColumn = foreignColumns[i];
					data.add(new Object[] { pTable.getName(),
							pColumn.getName(), fTable.getName(),
							fColumn.getName() });
				}
			}
			dataSet = new InMemoryDataSet(
					MetaModelHelper.createSelectItems(table.getColumns()), data);
		} else {
			throw new IllegalArgumentException(
					"Cannot materialize non information_schema table: " + table);
		}

		// Handle column subset
		dataSet = MetaModelHelper.getSelection(
				MetaModelHelper.createSelectItems(columns), dataSet);

		// Handle maxRows
		if (maxRows != -1) {
			dataSet = new MaxRowsDataSet(dataSet, maxRows);
		}
		return dataSet;
	}

	protected Schema getMainSchemaInternal() {
		if (_mainSchema == null) {
			_mainSchema = getMainSchema();
		}
		return _mainSchema;
	}

	/**
	 * @return the main schema that subclasses of this class produce
	 */
	protected abstract Schema getMainSchema() throws MetaModelException;

	/**
	 * @return the name of the main schema that subclasses of this class produce
	 */
	protected abstract String getMainSchemaName() throws MetaModelException;

	/**
	 * Executes a simple one-table query against a table in the main schema of
	 * the subclasses of this class.
	 * 
	 * @param table
	 *            the table to query
	 * @param columns
	 *            the columns to include
	 * @param maxRows
	 *            the maximum amount of rows needed or -1 if all rows are
	 *            wanted.
	 * @return a dataset with the raw table content of the colums
	 */
	protected abstract DataSet materializeMainSchemaTable(Table table,
			Column[] columns, int maxRows);
}